
; Program ...: DBIO.ASM
; Author ....: Evan M. Simpson
; Date ......: September 16, 1987
; Version ...: dBASE III PLUS or FoxBase+
;
; CALL DBIO WITH "@filename" to load in a text data file
;
; CALL DBIO WITH "tag1"+"tag2"+"tag3"+... to display the specified data
; CALL DBIO WITH ">"+"tag1"+"tag2"+... to save the screen first
; CALL DBIO WITH "<"+"tag1"+"tag2"+... to restore the screen first
;
;****************************************************************************
OPEN	EQU	3Dh		; Open a file.
CLOSE	EQU	3Eh		; Close a file.
READ	EQU	3Fh		; Read from a file.
EOL	EQU	100h		; End of Line.
EOF	EQU	200h		; End of File.
MONADDR	EQU	0B000h		; Segment address of monochrome screen memory.
COLADDR	EQU	0B800h		; Segment address of color screen memory.
;---------------------------------------
SAVE	MACRO	R1,R2,R3,R4,R5,R6,R7,R8
;
; This macro pushes a list of registers onto the stack in the order given.
;
	IRP	X,<R1,R2,R3,R4,R5,R6,R7,R8>
		IFNB    <X>
			PUSH X  
		ENDIF
	ENDM
	ENDM
;---------------------------------------
RESTORE	MACRO	R1,R2,R3,R4,R5,R6,R7,R8
;
; This macro pops a list of registers from the stack in reverse order.
;
	IRP	X,<R8,R7,R6,R5,R4,R3,R2,R1>
		IFNB    <X>
			POP X
		ENDIF
	ENDM
	ENDM
;---------------------------------------
DOS	MACRO	ROUT
	XOR	AX,AX
	MOV	AH,ROUT
	INT	21H
	ENDM
;---------------------------------------
JIFZ	MACRO	T1,TO
	TEST	T1,T1
	JE	TO
	ENDM
;---------------------------------------
JIFNZ	MACRO	T1,TO
	TEST	T1,T1
	JNE	TO
	ENDM
;---------------------------------------
ZERO	MACRO	Z1
	XOR	Z1,Z1
	ENDM
;****************************************************************************
CODE	SEGMENT	BYTE PUBLIC	'CODE'  
;---------------------------------------
DBIO	PROC	FAR
	ASSUME	CS:code,DS:code


START:	JMP	ENTRY
;===========================================================================
	DB	13,'DBIO.BIN v1.2 Copyright (c)1987',13,10
	DB	"Author ....: Evan M. Simpson",13,10
	DB	"Date ......: September 16, 1987",13,10
	DB	"Version ...: dBASE III PLUS or FoxBase+",13,10
	DB	13,10
	DB	"Thanks to Kenneth N. Getz for the proper method of reading/writing to CGA",13,10
	DB	13,10
	DB	"    DBIO.BIN is intended to allow easy storage of a program's output in a",13,10
	DB	"single, easily modified text file, and to display this output as quickly as",13,10
	DB	"possible. Two sorts of CALL's are allowed. Note that in neither is it",13,10
	DB	"required that parameters be placed in a memory variable prior to being",13,10
	DB	"passed. The first type is the command to read a data file into memory, as",13,10
	DB	"follows:",13,10
	DB	13,10
	DB	' CALL DBIO WITH "@filename"',13,10
	DB	13,10
	DB	"    The data file contains a set of `screens', each consisting of a two",13,10
	DB	"letter tag and zero or more `lines' and/or `boxes'. A line is a string of",13,10
	DB	"characters with a row,column position. A box is a bordered rectangular",13,10
	DB	"screen area.  DBIO interprets data files in this manner: The first character",13,10
	DB	"of each line is examined. If it is a capital letter, DBIO attempts to read a",13,10
	DB	"tag.  If it is a digit, DBIO attempts to read a 'line'. A `#' character",13,10
	DB	"means a box. All lines and boxes occurring after a tag belong to that tag.",13,10
	DB	"The standard format is:",13,10
	DB	13,10
	DB	" TAG: Two uppercase letters. All other characters on that line are ignored.",13,10
	DB	" Ex. 'HS','Help Screen', and 'H1*(S..' all produce the tag 'HS'",13,10
	DB	13,10
	DB	" LINE: Row number, Column number, and text separated by any non-digit.",13,10
	DB	" Ex. '10,2:Hi!','10 2 Hi!', and '10.2-Hi!' all mean 'Hi!' at row 10, col 2",13,10
	DB	13,10
	DB	" BOX: '#' followed by seven numbers: Top, Left, Bottom, Right, Inside Attr.,",13,10
	DB	"      Border Attr., and Border character all separated by non-digits.",13,10
	DB	" Ex. '#5,5,20,75,7,24,70' and '#5/5.20/75 7 24*70 main window' both mean a",13,10
	DB	"     box with corners (5,5) and (20,75) filled with white on black spaces",13,10
	DB	"     and a border of bright red 'E's. Note that if the inside attr. is zero,",13,10
	DB	"     the box is not filled. If the border char. is one of '', the border",13,10
	DB	"     will be drawn with appropriate symbols.",13,10
	DB	13,10
	DB	"    Once a data file has been loaded, DBIO can then be told to display one",13,10
	DB	"or more screens. This is accomplished by:",13,10
	DB	13,10
	DB	' CALL DBIO WITH "tag1"+"tag2"+"tag3"+...',13,10
	DB	13,10
	DB	'   Calling with "MM OR,w3PP" for example, would cause DBIO to display the',13,10
	DB	"screens with tags MM,OR, and PP. If a call to DBIO does not have `@' as its",13,10
	DB	"first character, DBIO will scour it for uppercase letters.",13,10
	DB	13,10
	DB	"   Display is accomplished by copying the screen to a buffer, writing all",13,10
	DB	"requested lines and boxes to this buffer, then writing the buffer back to",13,10
	DB	"screen memory. This results in near-instantaneous appearance of output.  If",13,10
	DB	"the first character of this sort of call is a '>', the screen will be saved",13,10
	DB	"before it is altered. If it is '<', the screen will be restored first.",13,10
	DB	26
BSET1	DB	'ɻȼͺ'
BSET2	DB	'ڿĳ'
BSET3	DB	'ոԾͳ'
BSET4	DB	'ַӽĺ'
SCREEN  DW	COLADDR			; Assume color
HANDLE  DW	0                       ; Storage for file handle
LASTCH	DW	0			; End of buffer pointer
TAGLOC	DW	0			; Tag pointer
BTOP	DB	0			; Box description array
BLEFT	DB	0
BBOT	DB	0
BRIGHT	DB	0
BIN	DB	0
BOUT	DB	0
BBORD	DB	6 DUP (0)		; Doubles as BSET0
INNER	DW	' ',0			; Temporary storage for box fill word
TEN	DB	10
LINELEN	DB	160			; Length of a single line in bytes.
OUTPUT	DB	8000 DUP(0)		; Output save area
BUFFER  DB	4000 DUP(0)		; Screen save area
STORAGE	DB	4000 DUP(0)		; Screen storage area
;===========================================================================
ENTRY:	SAVE	AX,BX,CX,DX,SI,DI,DS	; Save the environment
	PUSH	BX
	MOV	CS:SCREEN,COLADDR	; Get the screen mode (Color or Mono)
	MOV	AH,0Fh
	MOV	BH,0
	INT	10h
	CMP	AL,7
	JNE	ACT
	MOV	CS:SCREEN,MONADDR
ACT:	POP	BX
	CMP	BYTE PTR [BX],'@'	; See if we're reading or printing
	JNE	DISP
	MOV	DX,BX
	INC	DX			; We are reading. Point DX:DS to
	ZERO	CX			; the filename and call DOS
	DOS	OPEN
	JC	XDBIO			; Exit if there is an error
	MOV	DX,CS
	MOV	DS,DX			; Use CS for data from now on
	MOV	ES,DX
	MOV	HANDLE,AX		; Save handle
	CALL	PARSE			; Parse file to internal form
	MOV	BX,HANDLE		; and close the file
	DOS	CLOSE
	JMP	XDBIO
DISP:	CALL	GETSCR			; We are writing. Copy in the screen
	MOV	SI,BX
	ZERO	BX
	CMP	BYTE PTR [SI],'>'
	JNE	DISP1
STORE:	MOV	AL,CS:BYTE PTR [BX+OFFSET BUFFER]
	MOV	CS:BYTE PTR [BX+OFFSET STORAGE],AL
	INC	BX
	CMP	BX,4000
	JNE	STORE
DISP1:	CMP	BYTE PTR [SI],'<'
	JNE	SCR1
PULL:	MOV	AL,CS:BYTE PTR [BX+OFFSET STORAGE]
	MOV	CS:BYTE PTR [BX+OFFSET BUFFER],AL
	INC	BX
	CMP	BX,4000
	JNE	PULL
SCR1:	ZERO	AX
SCR2:	MOV	AH,AL
TAGGET:	LODSB				; Get a (hopefully) letter
	JIFZ	AL,TAGX			; if null, we're out of string
	CMP	AL,'A'
	JB	TAGGET
	CMP	AL,'Z'
	JA	TAGGET			; loop until it's an uppercase letter
	JIFZ	AH,SCR2			; if it's the first letter, get another
	CALL	SHOW			; find and write the screen specified
	JMP	SCR1			; then look for another tag
TAGX:	CALL	PUTSCR			; Write the screen out when all done
XDBIO:	RESTORE AX,BX,CX,DX,SI,DI,DS	; Restore the environment,
	RET				; And return.
DBIO	ENDP
;---------------------------------------
PARSE	PROC	NEAR
;
; This procedure takes the input file and converts it into tags, lines, and
; boxes in internal form.
;
	CLD
	CALL	READFL			; Read the first chunk of file
	MOV	DI,OFFSET OUTPUT	; DI points to the output area,
	MOV	TAGLOC,0		; and TAGLOC starts out null.
FIRST:	CALL	GETCHAR			; Get a character, discarding EOL's
FIRST1:	CMP	AX,EOL
	JE	FIRST
	CMP	AX,EOF
	JNE	CHTEST
	RET
CHTEST:	CMP	AL,'A'			; If the character is not in A-Z, see
	JB	NEWLN			; if it is a line
	CMP	AL,'Z'
	JA	NEWLN
	MOV	BH,AL			; It is a tag, so save the letter
NEWTAG:	CALL	GETCHAR			; Scan for a second letter
	JIFNZ	AH,FIRST1
	CMP	AL,'A'
	JB	NEWTAG
	CMP	AL,'Z'
	JA	NEWTAG
	MOV	AH,BH
	MOV	BX,TAGLOC
	MOV	TAGLOC,DI
	STOSW				; Once the tag is complete, write it
	MOV	AX,BX
	STOSW				; and the pointer to the previous tag
TODUMP:	JMP	DUMP
NEWLN:	CMP	AL,'0'			; If the char is not a digit, see if
	JB	NEWBOX			; it is a box
	CMP	AL,'9'
	JA	NEWBOX
	CALL	GETVAL			; It is a line, so get the row #
	JIFNZ	AH,EOST
	MOV	BTOP,DL
	CALL	GETCHAR			; Look for another number
	JIFNZ	AH,EOST
	CMP	AL,'0'			; If it is not there, scrap the line
	JB	TODUMP
	CMP	AL,'9'
	JA	TODUMP
	CALL	GETVAL			; Otherwise get the col #
	JIFNZ	AH,EOST
	MOV	AL,BTOP
	CMP	AL,24			; Make sure the numbers are in range
	JA	TODUMP
	CMP	DL,79
	JA	TODUMP
	OR	AL,128			; Add the line ID bits to them
	STOSB				; Write them to the output area
	MOV	AL,DL
	STOSB
TEXT:	CALL	GETCHAR			; Followed by the rest of the line
	STOSB
	JIFZ	AH,TEXT
EOST:	JMP	FIRST1			; Special stepping stone to FIRST1
NEWBOX:	CMP	AL,'#'			; If the char isn't '#', scrap the line
	JNE	DUMP
	MOV	BX,OFFSET BTOP		; It is a box
BOXDEF:	CALL	GETCHAR			; So try to get 7 numbers
	CMP	AL,'0'
	JB	ENDBOX
	CMP	AL,'9'
	JA	ENDBOX
	CALL	GETVAL
	CMP	BX,OFFSET BBORD
	JE	OKBOX
	JIFNZ	AH,EOST
	MOV	BYTE PTR [BX],DL
	INC	BX
	JMP	BOXDEF
OKBOX:	MOV	BYTE PTR [BX],DL	; We have 7 numbers, so make sure that
	PUSH	AX
	MOV	DH,BBOT
	MOV	DL,BRIGHT
	CMP	BTOP,DH			; Top<=Bottom
	JA	ENDBOX
	CMP	BLEFT,DL		; Left<=Right
	JA	ENDBOX
	CMP	DH,24			; Bottom<=24
	JA	ENDBOX
	CMP	DL,79			; Right<=79
	JA	ENDBOX
	OR	BTOP,192		; They are, so add the box ID bits
	MOV	CX,7
	MOV	BX,OFFSET BTOP		; And write the numbers
WBOX:	MOV	AL,BYTE PTR [BX]
	INC	BX
	STOSB
	LOOP	WBOX
ENDBOX:	POP	AX
	JMP	TOEND			; Terminate the line properly
DUMP:	CALL	GETCHAR			; Dump the rest of the line
TOEND:	JIFZ	AH,DUMP
	JMP	FIRST1			; and start over
PARSE	ENDP
;---------------------------------------
GETCHAR	PROC	NEAR
;
; This procedure reads characters from a file one at a time. It uses the screen
; buffer for input, and translates CR,LF into EOL. SI contains the offset from
; the screen buffer start.
;
	ZERO	AX
	CMP	LASTCH,0		; If the buffer is empty, return EOF
	JE	ISEOF
	MOV	AL,[SI+OFFSET BUFFER]	; Get the next character
	INC	SI
	CMP	SI,LASTCH		; If it is the last in the buffer,
	JNE	NOREAD			; read in more of the file.
	CALL	READFL
NOREAD:	CMP	AL,13			; If it is a CR, it may be EOL
	JE	ISEOL
	CMP	AL,26			; If it is ^Z, return EOF
	JNE	XGETCH
	MOV	LASTCH,0
ISEOF:	MOV	AX,EOF
	RET
ISEOL:	CMP	LASTCH,0		; If CR is not just before EOF,
	JE	XGETCH
	CMP	[SI+OFFSET BUFFER],10	; and is followed by a LF,
	JNE	XGETCH
	MOV	AX,EOL			; return EOL and dump the LF.
	INC	SI
	CMP	SI,LASTCH
	JE	READFL
XGETCH:	RET
GETCHAR	ENDP
;---------------------------------------
READFL	PROC	NEAR
;
; This procedure reads 4000 bytes of a file into BUFFER
;
	SAVE	AX,BX,CX,DX
	MOV	BX,HANDLE		; Retrieve the file handle
	MOV	CX,4000			; Try to read 4000 bytes
	MOV	DX,OFFSET BUFFER	; into the buffer.
	ZERO	SI
	MOV	LASTCH,SI		; Zero LASTCH and SI
	DOS	READ			; Do the read
	JC	XREADFL			; If an error occurs, leave LASTCH=0
	MOV	LASTCH,AX		; Otherwise LASTCH=# of bytes read
XREADFL:RESTORE	AX,BX,CX,DX
	RET
READFL	ENDP
;---------------------------------------
GETVAL	PROC	NEAR
;
; This procedure returns the numeric value of a string integer in DL
;
	ZERO	DX
VALLP:	SUB	AL,'0'			; Convert a digit to its value
	XCHG	AL,DL
	MUL	TEN			; N=N*10+ the new digit
	ADD	AL,DL
	XCHG	AL,DL
	CALL	GETCHAR			; Look for another digit
	CMP	AL,'0'
	JB	XGVAL
	CMP	AL,'9'
	JNA	VALLP
XGVAL:	RET
GETVAL	ENDP
;---------------------------------------
WRLINE	PROC	NEAR
;
; This procedure puts a line of text to buffer from the output area
;
	LODSB				; Get the row #
	AND	AX,63			; Eliminate the ID bits
	MUL	LINELEN
	MOV	DI,AX
	ZERO	AX
	LODSB				; Get the col #
	SHL	AL,1
	ADD	DI,AX
	ADD	DI,OFFSET BUFFER	; DI=ROW*80+COL*2+OFFSET BUFFER
	JMP	DNEXT
DTEXT:	STOSB				; Copy the text up to the null
	INC	DI			; Skipping attribute bytes
DNEXT:	LODSB
	JIFNZ	AL,DTEXT
	RET
WRLINE	ENDP
;---------------------------------------
WRBOX	PROC	NEAR
;
; This procedure puts a box to buffer from the output area
;
	MOV	DI,OFFSET BTOP
	MOV	CX,6
	REP	MOVSB			; Copy the first six box parameters
	LODSB
	MOV	CX,6
	REP	STOSB			; Write the border value into BBORD[]
	AND	BTOP,63			; Remove the ID bits from BTOP
	MOV	BX,OFFSET BBORD		; Default to a box made of BBORD's
	MOV	AL,BBORD
	CMP	AL,BSET1		; If BBORD is one of the special box
	JNE	BORD2			; corners, point BX to the set
	MOV	BX,OFFSET BSET1
BORD2:	CMP	AL,BSET2
	JNE	BORD3
	MOV	BX,OFFSET BSET2
BORD3:	CMP	AL,BSET3
	JNE	BORD4
	MOV	BX,OFFSET BSET3
BORD4:	CMP	AL,BSET4
	JNE	ADJUST
	MOV	BX,OFFSET BSET4
ADJUST:	ZERO	AH
	MOV	AL,BTOP
	MUL	LINELEN
	MOV	DI,AX
	ZERO	AX
	MOV	AL,BLEFT
	SHL	AL,1
	ADD	DI,AX
	ADD	DI,OFFSET BUFFER	; DI=BTOP*80+BLEFT*2+OFFSET BUFFER
	MOV	AH,BOUT			; AH=Border attribute
	ZERO	CX
	MOV	DH,BBOT
	SUB	DH,BTOP			; DH= the height of the box
	MOV	DL,BRIGHT
	SUB	DL,BLEFT		; DL= the width
VLIN:	JNE	HLIN			; Branch if width not zero
	MOV	AL,BYTE PTR [BX+5]	; Get vertical bar character
	MOV	CL,DH
	INC	CL
DVLIN:	MOV	WORD PTR [DI],AX	; Draw a vertical line
	ADD	DI,160
	LOOP	DVLIN
	JMP	XWBOX	
HLIN:	JIFNZ	DH,SQR			; Branch if not a horizontal line
	MOV	AL,BYTE PTR [BX+4]	; Get horizontal bar character
	MOV	CL,DL
	INC	CL
	REP	STOSW			; Draw a horizontal line
	JMP	XWBOX
SQR:	MOV	BP,DI			; Save DI for later use
	MOV	AL,BYTE PTR [BX]
	STOSW				; Get and write upper-left corner
	DEC	DL			; Make DL space between sides
	MOV	CL,DL
	JCXZ	UR			; Write top line
	MOV	AL,BYTE PTR [BX+4]
	REP	STOSW
UR:	MOV	AL,BYTE PTR [BX+1]	; Get and write upper-right corner
	STOSW
	MOV	AL,BYTE PTR [BX+5]	; Get the side and center chars
	XCHG	AX,INNER
	MOV	AH,BIN
	XCHG	AX,INNER
ROWS:	ADD	BP,160
	MOV	DI,BP
	DEC	DH
	JE	BOTTOM
	STOSW				; Write left side	
	XCHG	AX,INNER
	MOV	CL,DL
	JIFZ	AH,NOMID		; If BIN is zero, skip the middle
	REP	STOSW			; Write middle
	JMP	RIGHT
NOMID:	ADD	DI,CX
	ADD	DI,CX
RIGHT:	XCHG	AX,INNER
	STOSW				; Write right side
	JMP	ROWS
BOTTOM:	MOV	AL,BYTE PTR [BX+2]	; Get and write lower-left corner
	STOSW
	MOV	CL,DL
	JCXZ	BR
	MOV	AL,BYTE PTR [BX+4]
	REP	STOSW			; Write bottom line
BR:	MOV	AL,BYTE PTR [BX+3]	; Get and write lower-right corner
	STOSW
XWBOX:	RET
WRBOX	ENDP
;---------------------------------------
SHOW	PROC	NEAR
	SAVE	SI,DS
	MOV	DX,CS			; Point both DS and ES to CS
	MOV	DS,DX
	MOV	ES,DX
	CLD
	MOV	SI,TAGLOC		; Get the location of the last tag
LOOK:	JIFZ	SI,XSHOW		; If we're out of tags, exit
	CMP	AX,WORD PTR [SI]
	JE	MATCH			; If this isn't the tag we want,
	MOV	SI,WORD PTR [SI+2]	; get the pointer to the next tag
	JMP	LOOK
MATCH:	ADD	SI,4			; Skip the header, then get a byte
SENDS:	MOV	AL,BYTE PTR [SI]
	AND	AL,192			; Mask for the ID bits
	CMP	AL,128			; If they are 10xxxxxx, write a line
	JNE	NLINE
	CALL	WRLINE
	JMP	SENDS
NLINE:	CMP	AL,192			; If they are 11xxxxxx, write a box
	JNE	XSHOW			; Otherwise, that's all
	CALL	WRBOX
	JMP	SENDS
XSHOW:	RESTORE	SI,DS
	RET	
SHOW	ENDP
;---------------------------------------
GETSCR	PROC	NEAR
;
; This procedure copies the screen memory to a buffer.
; This procedure's timing is very tight, and it was written to minimize screen
; "snow."  It works correctly on B&W, Color, and EGA graphics cards that have
; been tested.
;
	SAVE	BX,DS
	PUSH	CS
	POP	ES			; Make CS and ES point to same segment.
	MOV	DI,OFFSET BUFFER	; Point DI to beginning of buffer.
	MOV	DS,CS:SCREEN		; Point DS to screen segment.
	ZERO	SI			; Point SI to beginning of screen.
	MOV	DX,3DAh			; Point DX to CGA status port.
	MOV	CX,2000			; Move length value into CX.
	CLD				; Set string direction to forward.
	CMP	CS:SCREEN,MONADDR	; See if we have monochrome
	JNE	W1
	REP	MOVSW			; If so, move words the easy way
	JMP	XGETS
W1:	IN	AL,DX			; Get 6845 status.
	RCR	AL,1			; Check horizontal retrace.
	JB	W1			; Loop if in horizontal retrace.
	CLI				; No interrupt during critical section.
W2:	IN	AL,DX			; Get 6845 status.
	RCR	AL,1			; Check for horizontal retrace.
	JNB	W2			; Loop if not in retrace.
	LODSW				; Load the video word.
	STI				; Allow interrupts.
	STOSW				; Store the video word.
	LOOP	W1			; Go to next word.
XGETS:	RESTORE BX,DS
	RET
GETSCR	ENDP
;---------------------------------------
PUTSCR	PROC	NEAR
;
; This procedure copies the buffer to the screen
;
	PUSH	CS
	POP	DS			; Make DS point to code segment
	MOV	SI,OFFSET BUFFER	; Point SI to beginning of buffer.
	MOV	ES,CS:SCREEN		; Point ES to screen memory.
	ZERO	DI			; Point DI to beginning of screen.
	MOV	CX,2000			; Get the number of words to show.
	CLD				; Clear the direction flag.
	CMP	CS:SCREEN,MONADDR	; See if GetMode found a mono screen.
	JNE	CLR			; If not, go to color branch.
	REP	MOVSW			; Move the correct number of words.
	RET				; And close out.
CLR:	MOV	DX,3DAh			; Point DX to CGA status port.
R1:	LODSW				; Get a video word.
	MOV	BP,AX			; Store it in BP.
	MOV	AH,9			; Get horiz and vert retrace mask.
R2:	IN	AL,DX			; Get 6845 status.
	RCR	AL,1			; Check horizontal retrace.
	JB	R2			; Loop if in horiz retrace.
	CLI				; No ints during critical section.
R3:	IN	AL,DX			; Get 6845 status.
	AND	AL,AH			; Check for both kinds of retrace.
	JNZ	R3			; Loop if not equal zero.
	MOV	AX,BP			; Get the video word.
	STOSW				; Store the video word.
	STI				; Allow interrupts.
	LOOP	R1			; Go do next word.
	RET
PUTSCR	ENDP
;---------------------------------------
CODE	ENDS
	END	START
